package org.enmas.client

import org.enmas.pomdp._, org.enmas.messaging._,
       org.enmas.util.FileUtils._, org.enmas.util.voodoo.ClassLoaderUtils._,
       scala.collection.immutable._, java.io.File,
       akka.actor._, akka.actor.Actor._, akka.dispatch._, akka.pattern.ask,
       akka.util.Timeout,
       scala.concurrent.duration._,
       com.typesafe.config.ConfigFactory

class ClientManager extends Actor with Provisionable {
  import ClientManager._, context._

  private var sessions = List[ActiveSession]()
  private var POMDPs = Set[LocallyAvailablePOMDP]()
  loadUserPOMDPs

  private def loadUserPOMDPs { if (pomdpDir.isDirectory) {
    for (f  ← pomdpDir.listFiles) if (f.getName.endsWith(".jar")) {
      POMDPs ++= loadPOMDPsFromFile(f)
    }
  }}

  private def loadPOMDPsFromFile(jar: File): List[LocallyAvailablePOMDP] = {
    findSubclasses[POMDP](jar).filterNot { _.getName contains "$"} map {
      clazz => LocallyAvailablePOMDP(clazz.newInstance, jar)
    }
  }

  private def scanHost(address: String, replyTo: ActorRef) {
    val host = actorFor(
      "akka://enmasServer@"+address+":"+serverPort+"/user/serverManager"
    )
    host ! RequestProvisions
    (host ? Discovery) onSuccess {
      case reply: DiscoveryReply => replyTo ! reply
    }
  }

  private def createServer(address: String, pomdpClassName: String) {
    POMDPs.find(_.pomdp.getClass.getName == pomdpClassName) match {
      case Some(availablePOMDP) => {
        readFile(availablePOMDP.jar) match {
          case Some(fileData) => {
            val host = actorFor(
              "akka://enmasServer@"+address+":"+serverPort+"/user/serverManager"
            )
            host ! Provision(fileData)
            host ! CreateServerFor(pomdpClassName)
          }
          case None => ()
        }
      }
      case None => ()
    }
  }

  private def createSession(server: ServerSpec) {
    val replyTo = sender
    POMDPs.find( _.pomdp.getClass.getName == server.pomdpClassName) match {
      case Some(availablePOMDP) => {
        val sessionRef = actorOf(Props(new Session(server.ref, availablePOMDP.pomdp)))
        watch(sessionRef) // subscribe to Terminated(sessionRef)
        val futureResult = (sessionRef ? 'Init)
        futureResult onSuccess {
          case e: Either[_,_] => e match {
            case Left(obj) => obj match { case id: Int => {
              sessions = (ActiveSession(sessionRef, id, server) :: sessions)
              replyTo ! true
            }}
            case _ => ()
          }
        }
        futureResult onFailure { case _ => replyTo ! false }
      }
      case None => replyTo ! false
    }
  }

  def receive = {

    case LoadPOMDPsFromFile(f) => POMDPs ++= loadPOMDPsFromFile(f)

    case GetLocalPOMDPs => sender ! POMDPList(POMDPs.toList map { _.pomdp })

    case Provision(fileData: FileData) => {
      val jarOption = provision[POMDP](fileData)
      jarOption map { jar => { POMDPs ++= loadPOMDPsFromFile(jar) }}
    }

    case ScanHost(serverHost) => scanHost(serverHost, sender)

    case CreateServer(serverHost, pomdpClassName) =>
      createServer(serverHost, pomdpClassName)

    case e: Error => e.cause.printStackTrace

    case m: CreateSession => createSession(m.server)

    case GetSessions => sender ! ActiveSessionList(sessions)

    case Terminated(deceasedActor) => {
      println("Received notice of some dead session")
      sessions.find(_.ref == deceasedActor) match { case Some(deadSession) => {
          sessions = sessions filterNot { _ == deadSession }
          unwatch(deadSession.ref)
        }
        case None => ()
      }
    }

    case error: Throwable => { println(
      "Error received from [%s]:\n%s".format(sender, error.getMessage)
    )}

    case _ => () // ignore unhandled messages
  }
}

object ClientManager extends App {
  import org.enmas.client.gui._, java.io.File

  // CM specific messages
  sealed case class ScanHost(serverHost: String)
  sealed case class CreateSession(server: ServerSpec)
  sealed case class LoadPOMDPsFromFile(file: File)
  case object GetSessions
  case object GetLocalPOMDPs
  sealed case class CreateServer(serverHost: String, pomdpClassName: String)

  // tuple representing an active session
  sealed case class ActiveSession(ref: ActorRef, id: Int, server: ServerSpec)

  // tuple representing a POMDP and the JAR file containing its bytecode
  sealed case class LocallyAvailablePOMDP(pomdp: POMDP, jar: File) {
    override def equals(obj: Any): Boolean = obj match {
      case other: LocallyAvailablePOMDP => pomdp.name == other.pomdp.name
      case _ => false
    }
  }

  sealed case class POMDPList(pomdps: List[POMDP])
  sealed case class ActiveSessionList(sessions: List[ActiveSession])

  // search locations for user-supplied code.
  val pomdpDir = new File("user/pomdp")
  val agentDir = new File("user/agent")
  val iterationSubscriberDir = new File("user/iterationSubscriber")

  implicit val timeout: Timeout = Timeout(3.seconds)
  val system = ActorSystem("enmasClient", ConfigFactory.load.getConfig("enmasClient"))
  val serverPort = 36627 // ENMAS
  val manager = system.actorOf(Props[ClientManager], "clientManager")
  val gui = new ClientGUI(manager)
  // val net = new NetInterface(manager)
}